package router

import (
    "os"
    "time"
    "syscall"
    "context"
    "strconv"
    "net/http"
    "os/signal"
    "path/filepath"
    "nexus-server/users"
    "nexus-server/logger"
    "nexus-server/config"
    "github.com/labstack/echo/v4"
    "github.com/labstack/echo/v4/middleware"
)

type Router struct {
    echo *echo.Echo
    version *RouterVersion
    token *RouterToken
    users *RouterUsers
    vms *RouterVMs
    tpl *RouterTemplates
}

var log = logger.GetInstance()
var conf = config.GetInstance()

func New() *Router {
    return &Router{
        echo.New(),
        NewRouterVersion(),
        NewRouterToken(),
        NewRouterUsers(),
        NewRouterVMs(),
        NewRouterTemplates()}
}

func (router *Router)Start(port int) {
    auth, err := NewAuth(users.GetUsersInstance())
    if err != nil {
        log.Fatal("Server error: "+err.Error())
    }

    router.echo.HideBanner = true
    router.echo.Use(middleware.BodyLimit(conf.Limits.MaxUploadSize))

    // Login management.
    router.echo.POST("/login", auth.Login)

    // Version management.
    router.echo.GET("/version", router.version.GetVersion)

    // Token management.
    tokenGroup := router.echo.Group("/token")
    tokenGroup.Use(middleware.JWTWithConfig(auth.GetTokenAccess()))
    tokenGroup.GET("/refresh", router.token.GetRefreshToken)

    // User management.
    usersGroup := router.echo.Group("/users")
    usersGroup.Use(middleware.JWTWithConfig(auth.GetTokenAccess()))
    usersGroup.GET("", router.users.GetUsers)
    usersGroup.GET("/whoami", router.users.GetLoggedUser)
    usersGroup.POST("", router.users.CreateUser)
    usersGroup.DELETE("/:email", router.users.DeleteUserByEmail)
    usersGroup.PUT("/:email/caps", router.users.SetUserCaps)
    usersGroup.PUT("/:email/unlock", router.users.UnlockUser)
    usersGroup.PUT("/:email/resetpwd", router.users.ResetPwd)
    usersGroup.PUT("/pwd", router.users.SetUserPwd)

    // VM management.
    vmsGroup := router.echo.Group("/vms")
    vmsGroup.Use(middleware.JWTWithConfig(auth.GetTokenAccess()))
    vmsGroup.GET("/:id", router.vms.GetListableVM)
    vmsGroup.GET("", router.vms.GetListableVMs)
    vmsGroup.GET("/:id/attach", router.vms.GetAttachableVM)
    vmsGroup.GET("/attach", router.vms.GetAttachableVMs)
    vmsGroup.GET("/del", router.vms.GetDeletableVMs)
    vmsGroup.GET("/start", router.vms.GetStartableVMs)
    vmsGroup.GET("/stop", router.vms.GetStoppableVMs)
    vmsGroup.GET("/reboot", router.vms.GetRebootableVMs)
    vmsGroup.GET("/edit", router.vms.GetEditableVMs)
    vmsGroup.GET("/editaccess", router.vms.GetModifiableVMAccessVMs)
    vmsGroup.GET("/exportdir", router.vms.GetDirExportableVMs)
    vmsGroup.GET("/importfiles", router.vms.GetFilesImportableVMs)

    vmsGroup.POST("", router.vms.CreateVM)
    vmsGroup.DELETE("/:id", router.vms.DeleteVM)
    vmsGroup.PUT("/:id", router.vms.EditVM)
    vmsGroup.PUT("/:id/start", router.vms.StartVM)
    vmsGroup.PUT("/:id/startwithcreds", router.vms.StartVMWithCreds)
    vmsGroup.POST("/:id/spicecreds", router.vms.VMSpiceCreds)
    vmsGroup.POST("/spicecreds", router.vms.VMSpiceCredsAny)
    vmsGroup.PUT("/:id/stop", router.vms.KillVM)
    vmsGroup.PUT("/:id/shutdown", router.vms.ShutdownVM)
    vmsGroup.PUT("/:id/reboot", router.vms.RebootVM)
    vmsGroup.PUT("/:id/access/:email", router.vms.SetVMAccessForUser)
    vmsGroup.DELETE("/:id/access/:email", router.vms.DeleteVMAccessForUser)
    vmsGroup.GET("/:id/exportdir", router.vms.ExportVMDir)
    vmsGroup.POST("/:id/importfiles", router.vms.ImportFilesToVM)

    // Template management.
    templatesGroup := router.echo.Group("/templates")
    templatesGroup.Use(middleware.JWTWithConfig(auth.GetTokenAccess()))
    templatesGroup.GET("/:id", router.tpl.GetTemplate)
    templatesGroup.GET("", router.tpl.GetTemplates)
    templatesGroup.POST("/vm", router.tpl.CreateTemplateFromVM)
    // Disabled route: uses huge amount of memory due to a bug in the Echo framework
    // templatesGroup.POST("/qcow", router.tpl.CreateTemplateFromQCOW)
    templatesGroup.GET("/:id/disk", router.tpl.ExportDisk)
    templatesGroup.DELETE("/:id", router.tpl.DeleteTemplate)
    templatesGroup.PUT("/:id", router.tpl.EditTemplate)

    // Starts server in a dedicated goroutine.
    go func() {
        certsDir := conf.CertsDir
        if err := router.echo.StartTLS(":"+strconv.Itoa(port), filepath.Join(certsDir, "/server-cert.pem"), filepath.Join(certsDir, "/server-key.pem")); err != nil {
            if err != http.ErrServerClosed {
                log.Fatal("Server error: "+err.Error())
            } else {
                log.Info("Shutting down server...")
            }
        }
    }()

    // Wait on dedicated signals.
    sigs := make(chan os.Signal, 1)
    signal.Notify(sigs, syscall.SIGINT, syscall.SIGTERM, syscall.SIGHUP)

    for {
        sig := <-sigs  // blocks on any of the above signals.
        log.Info("Server caught signal ("+sig.String()+")")
        // SIGHUP tells the server to reload its configuration.
        // This is just a placeholder: it does nothing for now.
        // TODO: reloads users and templates? (very tricky though)
        if sig == syscall.SIGHUP {
            log.Info("Reloading server configuration")
        } else {
            // Any other signal tells the server to gracefully shutdown.
            break
        }
    }

    ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
    defer cancel()
    if err := router.echo.Shutdown(ctx); err != nil {
        log.Fatal("Server shutdown failed: "+err.Error())
    }
    log.Info("Server shutdown completed.")
}
